from ._link import FreeSpaceLink, LogNormalLink
from ._sensor import SensorNode, RADIO_CONFIG
from ._network import SensorNetwork

__all__ = ['SensorNode', 'SensorNetwork', 'RADIO_CONFIG', 'FreeSpaceLink', 'LogNormalLink']


# coding: utf-8
#
# Copyright (C) 2020 wsn-toolkit
#
# This program was written by Edielson P. Frigieri <edielsonpf@gmail.com>

from wsntk.simulator import SimuNet
import numpy as np
import logging

#from scipy.spatial.distance import cdist

logging.basicConfig(format='%(asctime)-15s - %(message)s', level=logging.INFO)
logger = logging.getLogger("network simulation")

## set this to true if you want to plot node positions
DRAW = True

## number of sensors
nr_sensors = 10

## simulation area (km)
max_x, max_y = 15, 15

np.random.seed(0xffff)

## Free Space path loos-based simulation
#net = SimuNet(nr_sensors, dimensions=(max_x, max_y), loss = "FSPL", sigma = 8.7, gamma = 2.2, radio = "DEFAULT")

## Log-Normal path loos-based simulation
net = SimuNet(nr_sensors, dimensions=(max_x, max_y), loss = "LNPL", sigma = 8.7, gamma = 2.2, radio = "DEFAULT")

## Log-Normal path loos-based simulation with ESP32-WROOM-32U radio sensors
#net = SimuNet(nr_sensors, dimensions=(max_x, max_y), loss = "LNPL", sigma = 8.7, gamma = 2.2, radio = "ESP32-WROOM-32U")


if DRAW:
    import matplotlib.pyplot as plt
    
    # Create new Figure and an Axes which fills it.
    ax = plt.subplot(111)
    ax.set_xlim(0, max_x)
    ax.set_ylim(0, max_y)

    # Construct the scatter which will be update during simulation 
    scat_sensors = ax.scatter([], [], s=60, lw=0.5, zorder=2)
    lines = [ax.plot([],[], 'C3', zorder=1, lw=0.3)[0] for j in range(nr_sensors*nr_sensors)]
    
step = 0
for sensors,links in net:
    
    step = step+1
    if step%10 == 0: 
        logger.info('Step %s'% step)
        #print(sensors)
        #print(links)
        
    
    if DRAW:
        #cleanup all lines
        for line in lines:
            line.set_data([],[])
        
        #update only the lines which represents a valid link
        lnr = 0
        i = 0
        for link_set in links:
            j = 0
            for link in link_set:
                if link == 1:
                    lines[lnr].set_data([sensors[i,0],sensors[j,0]], [sensors[i,1],sensors[j,1]])
                    lnr += 1
                j = j + 1
            i = i + 1

        #update the sensors position
        scat_sensors.set_offsets(sensors)

        plt.draw()
        plt.pause(0.5)



# Author: Edielson P. Frigieri <edielsonpf@gmail.com>
#
# License: MIT

import pytest
import numpy as np

from wsntk import network
from wsntk.network import FreeSpaceLink, LogNormalLink

def test_free_space_link():
    # Test FreeSpace link creation with default values.
    
    link = FreeSpaceLink()
    
    assert link.loss(distance = 10, frequency = 933e6) == 111.83763287493002
    
def test_log_normal_link():
    # Test FreeSpace link creation with default values.
    np.random.seed(0xffff)

    link = LogNormalLink(sigma = 8.7, gamma = 2.2)
    
    assert link.loss(distance = 10, frequency = 933e6) ==  128.46425705711366


# coding: utf-8
#
# Copyright (C) 2020 wsn-toolkit
#
# This program was written by Edielson P. Frigieri <edielsonpf@gmail.com>

"""Path loss models."""

from abc import ABCMeta, abstractmethod
import numpy as np
import math

DEAFULT_CONSTANT = 32.44 #For distance and frequency in kilometers and megahertz.

class FreeSpaceLink(metaclass=ABCMeta):
    """Base class for path loss models."""
    
    def __init__(self):
        pass
    
    def _friss_loss(self, distance, frequency):
        """
        Calculate the link loss.
                
        Parameters
        ----------
        distance : double
           The distance between two nodes (Km).
        
        frequency: double
            The frequency of operation (Hz).  
        
        Returns
        -------
        double
            The Friss loss evaluated for `distance` and `frequency`.
        """
        return (DEAFULT_CONSTANT + 20*math.log10(frequency/1e6) + 20*math.log10(distance))

    def loss(self, distance, frequency):
        """
        Calculate the link loss.
        For distance and frequency in kilometers and megahertz, respectively, 
        the constant becomes 32.44
        
        Parameters
        ----------
        distance : double
           The distance between two nodes (Km).
        
        frequency: double
            The frequency of operation (Hz).  
        
        Returns
        -------
        double
            The loss evaluated for `distance` and `frequency`.
        """
        return self._friss_loss(frequency, distance)


class LogNormalLink(FreeSpaceLink):
    """Base class for path loss models."""
    
    def __init__(self, sigma = 8.7, gamma = 2.2):
        
        self.sigma = sigma
        self.gamma = gamma
        super(LogNormalLink, self).__init__()
        
    def loss(self, distance, frequency):
        """
        Calculate the link loss.
        For distance and frequency in kilometers and megahertz, respectively. 
        
        The log-normal path-loss model may be considered as a generalization of the free-space Friis equation
        where a random variable is added in order to account for shadowing (large–scale fading) effects.
        See:    https://www.sciencedirect.com/topics/computer-science/path-loss-model
                https://en.wikipedia.org/wiki/Log-distance_path_loss_model

                Building Type	            Frequency of Transmission	gamma 	sigma [dB]
                Vacuum, infinite space		                            2.0	    0
                Retail store	            914 MHz	                    2.2	    8.7
                Grocery store	            914 MHz	                    1.8	    5.2
                Office with hard partition	1.5 GHz	                    3.0	    7
                Office with soft partition	900 MHz	                    2.4	    9.6
                Office with soft partition	1.9 GHz	                    2.6	    14.1
                Textile or chemical	        1.3 GHz	                    2.0	    3.0
                Textile or chemical	        4 GHz	                    2.1	    7.0, 9.7
                Office	                    60 GHz	                    2.2	    3.92
                Commercial	                60 GHz	                    1.7	    7.9

        Parameters
        ----------
        distance : double
           The distance between two nodes (Km).
        
        frequency: double
            The frequency of operation (Hz).  
        
        Returns
        -------
        double
            The loss evaluated for `distance` and `frequency`.
        """
        return self._friss_loss(distance, frequency) + 10*self.gamma*math.log10(distance) + np.random.normal(0, self.sigma)

    


# coding: utf-8
#
# Copyright (C) 2020 wsn-toolkit
#
# This program was written by Edielson P. Frigieri <edielsonpf@gmail.com>

"""Sensor nodes for wirelesss sensor networks simulation."""

from abc import ABCMeta, abstractmethod
from numpy.random import rand

RADIO_CONFIG = {"DEFAULT":          {"min_tx_power": -15.0, "max_tx_power": 27.0, "rx_sensitivity": -80.0, "frequency": 933e6},
                "ESP32-WROOM-32U":  {"min_tx_power": -12.0, "max_tx_power": 9.0, "rx_sensitivity": -97.0, "frequency": 2.4e9}}

class BaseNode(metaclass=ABCMeta):
    """Base class for sensor node."""
    
    def __init__(self, dimensions):
        
        self.dimensions = dimensions
        self._init_node()

    def _init_node(self):
        ndim = len(self.dimensions)
        self.position = rand(ndim) * self.dimensions

    def set_position(self, position):
        """
        Set node position
        
        Parameters
        ----------
        position : tuple of Double
           The x and y position of the sensor.
        
        Returns
        -------
        No data returned
        """
        ndim = len(self.dimensions)

        if(ndim != len(position)):
            raise ValueError("Position lenght different then expected. Expected %s, received %s." %(ndim, len(position)))    
        
        for index in range(ndim):
            if (position[index] > self.dimensions[index]):
                raise ValueError("Position exceeded dimensions limits.")
        
        self.position = position    
        
    def get_position(self):
        """
        Get node position
        
        Parameters
        ----------
        No parameters.
        
        Returns
        -------
        Tuple of Double
           The current x and y position of the sensor
        """
        return self.position

    def _update_position(self):
        """ Update the sensor position based on mobility models """
        #ndim = len(self.position)
        #step = 0.1*rand(ndim)
        #self.position = self.position + step
        pass

    @abstractmethod
    def __iter__(self):
        """Used to return an iteractor from a node"""
        raise NotImplementedError
    
    @abstractmethod
    def __next__(self):
        """Interator next"""
        raise NotImplementedError

class SensorNode(BaseNode):
    """
    Sensor node class.
        
    Required arguments:
        
        *dimensions*:
        Tuple of doubles, the area limits of the sensor.
          
        *radio*:
        Enumerator <RADIO_CONFIG>, the radio type to be used in the sensor.
    """
    def __init__(self, dimensions, radio = "DEFAULT"):
        
        super(SensorNode, self).__init__(dimensions)
        #initialize radio configuration
        self._set_radio_config(radio)
        
    def _set_radio_config(self, radio_type):
        """ Collect the radio parameters used in the sensor """

        radio_params = self._get_radio_params(radio_type)
        for param in radio_params: 
            if param == "max_tx_power":
               self.max_tx_power = radio_params[param]
            elif param == "min_tx_power":
               self.min_tx_power = radio_params[param]
            elif param == "rx_sensitivity":
               self.rx_sensitivity = radio_params[param]
            elif param == "frequency":
               self.frequency = radio_params[param]
            else:
                raise ValueError("Radio parameter not expected: %s." %(param))
        
        #initialize radio with maximun tx_power                    
        self.tx_power = self.max_tx_power        

    def _get_radio_params(self, radio_type):
        """ Retrieve the radio parameters based on specified type """
        radio_type = str(radio_type).upper()
        try:
            return RADIO_CONFIG[radio_type]
        except KeyError as e:
            raise ValueError("Radio %s is not supported." % radio_type) from e

    def _update_energy(self):
        """ Update the sensor energy based on consumption models """
        pass
    
    def set_txpower(self, tx_power):
        """
        Set radio transmission power
        
        Parameters
        ----------
        tx_power : {double}
            Transmission power to be configured in the radio
        
        Returns
        -------
        No data returned
        """
        
        if((tx_power >= self.min_tx_power) and (tx_power <= self.max_tx_power)):
            self.tx_power = tx_power    
        else:
            raise ValueError("Parameter out of radio power specification. Expected value from %s dBm to %s dBm." %(self.min_tx_power, self.max_tx_power))
                        

    def get_txpower(self):
        """
        Get radio transmission power
        
        Parameters
        ----------
        No parameters
        
        Returns
        -------
        double number
            The current configured transmission power
        """
        return self.tx_power

    def get_rxsensitivity(self):
        """
        Get radio receiver sensitivity
        
        Parameters
        ----------
        No parameters
        
        Returns
        -------
        double number
            The current configured receiver sensitivity
        """
        return self.rx_sensitivity

    def get_frequency(self):
        """
        Get the radio frequency
        
        Parameters
        ----------
        No parameters
        
        Returns
        -------
        double number
            The current configured receiver sensitivity
        """
        return self.frequency

    def __iter__(self):
        """Interator"""
        return self
    
    def __next__(self):
        self._update_position()
        self._update_energy()
        return self.position
           





# coding: utf-8
#
# Copyright (C) 2020 wsn-toolkit
#
# This program was written by Edielson P. Frigieri <edielsonpf@gmail.com>

import pytest

from wsntk import network
from wsntk.network import SensorNode, RADIO_CONFIG

def test_sensor_node():
    # Test SensorNode creation with default values.
    
    sensor = SensorNode(dimensions = (10.0, 10.0), radio = "DEFAULT")
    
    parameter = RADIO_CONFIG["DEFAULT"] 
    assert sensor.get_txpower() == parameter["max_tx_power"]
    assert sensor.get_frequency() == parameter["frequency"]

def test_set_position():
    # Test setting the a new position.
    
    sensor = SensorNode(dimensions = (10.0, 10.0), radio = "DEFAULT")
    
    position = (0.5,0.5)
    sensor.set_position(position)

    assert sensor.get_position() == position

def test_set_position_wrong_len_raise_value_error():
    # Test setting the a new position with wrong len.
    
    sensor = SensorNode(dimensions = (10.0, 10.0), radio = "DEFAULT")
    
    error_msg = ('Position lenght different then expected. Expected 2, received 3.')
    
    with pytest.raises(ValueError, match=error_msg):
        sensor.set_position(position = (0.5,0.5,0.5))

def test_set_position_outside_limits_raise_value_error():
    # Test setting the a new position outside limits.
    sensor = SensorNode(dimensions = (10.0, 10.0), radio = "DEFAULT")
    
    error_msg = ('Position exceeded dimensions limits.')
    
    with pytest.raises(ValueError, match=error_msg):
        sensor.set_position(position = (11.0, 11.0))

def test_sensor_node_with_unkown_radio_raises_value_error():
    # Test SensorNode creation with a unknown radio type.
    
    error_msg = ('Radio UNKNOWN is not supported.')
        
    with pytest.raises(ValueError, match=error_msg):
        sensor = SensorNode(dimensions = (10.0, 10.0), radio = "UNKNOWN")

def test_set_tx_power_within_range():
    # Test setting the tx_power within expected range.
    
    sensor = SensorNode(dimensions = (10.0, 10.0), radio = "ESP32-WROOM-32U")
    
    parameters = RADIO_CONFIG["ESP32-WROOM-32U"]
    new_tx_power = (parameters["min_tx_power"] + parameters["max_tx_power"])/2
    sensor.set_txpower(new_tx_power)
    assert sensor.get_txpower() == new_tx_power


def test_set_tx_power_below_range_raises_value_error():
    # Test exception when setting the tx_power below expected range.
        
    error_msg = ('Parameter out of radio power specification. Expected value from -12.0 dBm to 9.0 dBm.')
        
    sensor = SensorNode(dimensions = (10.0, 10.0), radio = "ESP32-WROOM-32U")
    
    parameters = RADIO_CONFIG["ESP32-WROOM-32U"]
    with pytest.raises(ValueError, match=error_msg):
        sensor.set_txpower(parameters["min_tx_power"] - 13.0)


def test_set_tx_power_above_range_raises_value_error():
    # Test exception when setting the tx_power above expected range.
    
    error_msg = ('Parameter out of radio power specification. Expected value from -12.0 dBm to 9.0 dBm.')
        
    sensor = SensorNode(dimensions = (10.0, 10.0), radio = "ESP32-WROOM-32U")
    
    parameters = RADIO_CONFIG["ESP32-WROOM-32U"]
    with pytest.raises(ValueError, match=error_msg):
        sensor.set_txpower(parameters["max_tx_power"] + 10.0)

def test_get_rx_sensitivity():
    # Test getting the rx sensitivity.
    
    parameters = RADIO_CONFIG["ESP32-WROOM-32U"]

    sensor = SensorNode(dimensions = (10.0, 10.0), radio = "ESP32-WROOM-32U")
    assert sensor.get_rxsensitivity() == parameters["rx_sensitivity"]
    


# Authors: Edielson P. Frigieri <edielsonpf@gmail.com> (main author)
#
# License: MIT

"""Sensor networks simulation."""

from wsntk.network import SensorNode
from wsntk.network import FreeSpaceLink, LogNormalLink

from abc import ABCMeta, abstractmethod

from numpy.random import rand
import math
import numpy as np

class BaseNetwork(metaclass=ABCMeta):
    """
        Sensor network class.
        This class is implments a sensor network which involves elemts such as sensors and links.
        
        Required arguments:
        
          *nr_sensors*:
            Integer, the number of sensors.
          
          *dimensions*:
            Tuple of Integers, the x and y dimensions of the simulation area in kilometers.

          *loss function*
            String, the path loss function
          
          *sigma*
            Double, standard deviation applied to specific path loss models

          *gamma*
            Double, the path loss exponent

          *radio*
            String, the radio type usd on all sensors
    """

    link_models = {
        "FSPL": (FreeSpaceLink,),
        "LNPL": (LogNormalLink,),
    }

    def __init__(self, nr_sensors, dimensions, loss = "FSPL", sigma = 8.7, gamma = 2.2, radio = "DEFAULT"):
        
        self.nr_sensors = nr_sensors
        self.dimensions = dimensions
        self.radio = radio
        self.sigma = sigma
        self.gamma = gamma
        
        self.link = self._init_link(loss)
        self.sensors = self._init_sensors(nr_sensors, dimensions, radio)
            
    def _init_link(self, loss):
        """Get ``LinkClass`` object for str ``loss``. """
        try:
            link_ = self.link_models[loss]
            link_class, args = link_[0], link_[1:]
            if loss in ('LNPL'):
                args = (self.sigma, self.gamma)
            return link_class(*args)
        except KeyError as e:
            raise ValueError("The link loss %s is not supported. " % loss) from e


    def _init_sensors(self, nr_sensors, dimensions, radio):
        """Initializes the simulaiton creating all sensors with respective configuration. """                
        
        sensors = []
        #instantiate all sensors
        for i in range (nr_sensors):
            #instantiate a sensor
            sensor = SensorNode(dimensions, radio)
            #Add sensor to the network
            sensors.append(sensor)
        
        return sensors

    def _distance(self, pos_a, pos_b):
        """Calculate the euclidean distance between two positions"""
        return (math.sqrt(((pos_a[0]-pos_b[0])**2)+((pos_a[1]-pos_b[1])**2)))
    
    def __iter__(self):
        """Generator which returns the current links and sensors after update."""
        while True:
            sensors = self._update_sensors()
            links = self._update_links()
            yield sensors,links

    @abstractmethod
    def _update_sensors(self):
        """Update the sensors status: position, energy, etc."""
        raise NotImplementedError
    
    @abstractmethod
    def _update_links(self):
        """Update the links status based on the new sensors status."""
        raise NotImplementedError
       
    
class SensorNetwork(BaseNetwork):
    """
        Sensor network class.
        This class is implments a sensor network which involves elemts such as sensors and links.
        
        Required arguments:
        
          *nr_sensors*:
            Integer, the number of sensors.
          
          *dimensions*:
            Tuple of Integers, the x and y dimensions of the simulation area in kilometers.

          *loss function*
            String, the path loss function

          *sigma*
            Double, standard deviation applied to specific path loss models
          
          *gamma*
            Double, the path loss exponent

          *radio*
            String, the radio type usd on all sensors
    """
    def __init__(self, nr_sensors, dimensions, loss = "FSPL", sigma = 8.7, gamma = 2.2,  radio = "DEFAULT"):
        
        super(SensorNetwork, self).__init__(nr_sensors, dimensions, loss, sigma, gamma, radio)
    
    def _update_sensors(self):
        positions = np.empty((0, len(self.dimensions)))
        for sensor in self.sensors:
            position = next(iter(sensor))
            positions = np.append(positions, position.reshape(1,len(self.dimensions)), axis = 0)
            
        return positions
 
    def _link_status(self, rx_sensor, tx_sensor):
        
        if rx_sensor != tx_sensor:
            distance = self._distance(rx_sensor.get_position(), tx_sensor.get_position())
            loss = self.link.loss(distance, rx_sensor.frequency)
            rx_power = tx_sensor.tx_power - loss
            
            if(rx_power >= rx_sensor.rx_sensitivity):
                link_status = 1
            else:
                link_status = 0
        else:
            link_status = 0

        return link_status

    def _update_links(self):
        links = []
        for rx_sensor in self.sensors:
            aux_links = []
            for tx_sensor in self.sensors:
                    aux_links.append(self._link_status(rx_sensor, tx_sensor))
            links.append(aux_links)    
        
        return links
    



# coding: utf-8
#
# Copyright (C) 2020 wsn-toolkit
#
# This program was written by Edielson P. Frigieri <edielsonpf@gmail.com>

from wsntk.network import SensorNode

position = (0.5, 0.5)

#create two sensors with different radio configurations
Sensor1 = SensorNode(position) # node with default radio configuration
Sensor2 = SensorNode(position, radio = "ESP32-WROOM-32U")

#show current radio configuration
print("Sensor1: tx power =  %s, rx_sensitivity = %s " %(Sensor1.get_txpower(), Sensor1.get_rxsensitivity()))
print("Sensor2: tx power =  %s, rx_sensitivity = %s " %(Sensor2.get_txpower(), Sensor2.get_rxsensitivity()))

#set new tranmission power for sensor 1
Sensor1.set_txpower(8.0)
print("Sensor1: new tx power =  %s, rx_sensitivity = %s " %(Sensor1.get_txpower(), Sensor1.get_rxsensitivity()))

#set new position for sensor 2
print("Sensor2: old position =  (%s, %s) " %(Sensor2.get_position()))
Sensor2.set_position((-5.0, 5.0))
print("Sensor2: new position =  (%s, %s) " %(Sensor2.get_position()))




#! /usr/bin/env python
#
# Copyright (C) 2020 wsn-toolkit
#
# This program was written by Edielson P. Frigieri <edielsonpf@gmail.com>
#               
# License: MIT
#version = attr: wsntk.__version__


import setuptools

if __name__ == '__main__':
    setuptools.setup(
		setup_requires=['pytest-runner', 'pbr'],
		pbr=True,
		tests_require=['pytest'],
    )
	